package com.sun.jndi.dns;

import java.io.StreamCorruptedException;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.NoSuchElementException;
import java.util.Enumeration;
import javax.naming.CompositeName;
import java.util.Iterator;
import javax.naming.InvalidNameException;
import java.util.ArrayList;
import javax.naming.Name;

public final class DnsName implements Name
{
    private String domain;
    private ArrayList<String> labels;
    private short octets;
    private static final long serialVersionUID = 7040187611324710271L;
    
    public DnsName() {
        this.domain = "";
        this.labels = new ArrayList();
        this.octets = 1;
    }
    
    public DnsName(String s) throws InvalidNameException {
        this.domain = "";
        this.labels = new ArrayList();
        this.octets = 1;
        this.parse(s);
    }
    
    private DnsName(DnsName dnsName, int n, int n2) {
        this.domain = "";
        this.labels = new ArrayList();
        this.octets = 1;
        this.labels.addAll(dnsName.labels.subList(dnsName.size() - n2, dnsName.size() - n));
        if (this.size() == dnsName.size()) {
            this.domain = dnsName.domain;
            this.octets = dnsName.octets;
        }
        else {
            for (String s : this.labels) {
                if (s.length() > 0) {
                    this.octets += (short)(s.length() + 1);
                }
                else {
                    continue;
                }
            }
        }
    }
    
    @Override
    public String toString() {
        if (this.domain == null) {
            StringBuilder sb = new StringBuilder();
            for (String s : this.labels) {
                if (sb.length() > 0 || s.length() == 0) {
                    sb.append('.');
                }
                escape(sb, s);
            }
            this.domain = sb.toString();
        }
        return this.domain;
    }
    
    public boolean isHostName() {
        Iterator<String> iterator = this.labels.iterator();
        while (iterator.hasNext()) {
            if (!isHostNameLabel((String)iterator.next())) {
                return false;
            }
            continue;
        }
        return true;
    }
    
    public short getOctets() {
        return this.octets;
    }
    
    @Override
    public int size() {
        return this.labels.size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.size() == 0;
    }
    
    @Override
    public int hashCode() {
        int n = 0;
        for (int i = 0; i < this.size(); ++i) {
            n = 31 * n + this.getKey(i).hashCode();
        }
        return n;
    }
    
    @Override
    public boolean equals(Object o) {
        return o instanceof Name && !(o instanceof CompositeName) && this.size() == ((Name)o).size() && this.compareTo(o) == 0;
    }
    
    @Override
    public int compareTo(Object o) {
        return this.compareRange(0, this.size(), (Name)o);
    }
    
    @Override
    public boolean startsWith(Name name) {
        return this.size() >= name.size() && this.compareRange(0, name.size(), name) == 0;
    }
    
    @Override
    public boolean endsWith(Name name) {
        return this.size() >= name.size() && this.compareRange(this.size() - name.size(), this.size(), name) == 0;
    }
    
    @Override
    public String get(int n) {
        if (n < 0 || n >= this.size()) {
            throw new ArrayIndexOutOfBoundsException();
        }
        return (String)this.labels.get(this.size() - n - 1);
    }
    
    @Override
    public Enumeration<String> getAll() {
        return new Enumeration<String>() {
            int pos = 0;
            
            @Override
            public boolean hasMoreElements() {
                return this.pos < DnsName.this.size();
            }
            
            @Override
            public String nextElement() {
                if (this.pos < DnsName.this.size()) {
                    return DnsName.this.get(this.pos++);
                }
                throw new NoSuchElementException();
            }
        };
    }
    
    @Override
    public Name getPrefix(int n) {
        return new DnsName(this, 0, n);
    }
    
    @Override
    public Name getSuffix(int n) {
        return new DnsName(this, n, this.size());
    }
    
    @Override
    public Object clone() {
        return new DnsName(this, 0, this.size());
    }
    
    @Override
    public Object remove(int n) {
        if (n < 0 || n >= this.size()) {
            throw new ArrayIndexOutOfBoundsException();
        }
        String s = (String)this.labels.remove(this.size() - n - 1);
        int length = s.length();
        if (length > 0) {
            this.octets -= (short)(length + 1);
        }
        this.domain = null;
        return s;
    }
    
    @Override
    public Name add(String s) throws InvalidNameException {
        return this.add(this.size(), s);
    }
    
    @Override
    public Name add(int n, String element) throws InvalidNameException {
        if (n < 0 || n > this.size()) {
            throw new ArrayIndexOutOfBoundsException();
        }
        int length = element.length();
        if ((n > 0 && length == 0) || (n == 0 && this.hasRootLabel())) {
            throw new InvalidNameException("Empty label must be the last label in a domain name");
        }
        if (length > 0) {
            if (this.octets + length + 1 >= 256) {
                throw new InvalidNameException("Name too long");
            }
            this.octets += (short)(length + 1);
        }
        int index = this.size() - n;
        verifyLabel(element);
        this.labels.add(index, element);
        this.domain = null;
        return this;
    }
    
    @Override
    public Name addAll(Name name) throws InvalidNameException {
        return this.addAll(this.size(), name);
    }
    
    @Override
    public Name addAll(int n, Name name) throws InvalidNameException {
        if ((name instanceof DnsName)) {
            DnsName dnsName2 = (DnsName)name;
            if (dnsName2.isEmpty()) {
                return this;
            }
            if ((n > 0 && dnsName2.hasRootLabel()) || (n == 0 && this.hasRootLabel())) {
                throw new InvalidNameException("Empty label must be the last label in a domain name");
            }
            int octets = (short)(this.octets + dnsName2.octets - 1);
            if (octets > 255) {
                throw new InvalidNameException("Name too long");
            }
            this.octets = octets;
            this.labels.addAll(this.size() - n, dnsName2.labels);
            if (this.isEmpty()) {
                this.domain = dnsName2.domain;
            }
            else if (this.domain == null || dnsName2.domain == null) {
                this.domain = null;
            }
            else if (n == 0) {
                this.domain = this.domain + (dnsName2.domain.equals(".") ? "" : ".") + dnsName2.domain;
            }
            else if (n == this.size()) {
                this.domain = dnsName2.domain + (this.domain.equals(".") ? "" : ".") + this.domain;
            }
            else {
                this.domain = null;
            }
        }
        else if ((name instanceof CompositeName)) {
            DnsName dnsName = (DnsName)name;
        }
        else {
            for (int i = name.size() - 1; i >= 0; --i) {
                this.add(n, name.get(i));
            }
        }
        return this;
    }
    
    boolean hasRootLabel() {
        return !this.isEmpty() && this.get(0).equals("");
    }
    
    private int compareRange(int n, int n2, Name name) {
        if ((name instanceof CompositeName)) {
            name = (Name)name;
        }
        int min = Math.min(n2 - n, name.size());
        int i = 0;
        while (i < min) {
            String value = this.get(i + n);
            String value2 = name.get(i);
            int n3 = this.size() - (i + n) - 1;
            int compareLabels = compareLabels(value, value2);
            if (compareLabels != 0) {
                return compareLabels;
            }
            ++i;
        }
        return n2 - n - name.size();
    }
    
    String getKey(int n) {
        return keyForLabel(this.get(n));
    }
    
    private void parse(String domain) throws InvalidNameException {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < domain.length(); ++i) {
            int char1 = domain.charAt(i);
            if (char1 == '\\') {
                int escapedOctet = getEscapedOctet(domain, i++);
                if (isDigit(domain.charAt(i))) {
                    i += 2;
                }
                sb.append((char)escapedOctet);
            }
            else if (char1 != '.') {
                sb.append((char)char1);
            }
            else {
                this.add(0, sb.toString());
                sb.delete(0, i);
            }
        }
        if (!domain.equals("") && !domain.equals(".")) {
            this.add(0, sb.toString());
        }
        this.domain = domain;
    }
    
    private static char getEscapedOctet(String s, int n) throws InvalidNameException {
        try {
            int char1 = s.charAt(++n);
            if (isDigit((char)char1)) {
                int char2 = s.charAt(++n);
                int char3 = s.charAt(++n);
                if (isDigit((char)char2) && isDigit((char)char3)) {
                    return (char)((char1 - 48) * 100 + (char2 - 48) * 10 + (char3 - 48));
                }
                throw new InvalidNameException("Invalid escape sequence in " + s);
            }
            else {
                return char1;
            }
        }
        catch (IndexOutOfBoundsException ex) {
            throw new InvalidNameException("Invalid escape sequence in " + s);
        }
    }
    
    private static void verifyLabel(String s) throws InvalidNameException {
        if (s.length() > 63) {
            throw new InvalidNameException("Label exceeds 63 octets: " + s);
        }
        int i = 0;
        while (i < s.length()) {
            if ((s.charAt(i) & '\uff00') != 0x0) {
                throw new InvalidNameException("Label has two-byte char: " + s);
            }
            ++i;
        }
        return;
    }
    
    private static boolean isHostNameLabel(String s) {
        int i = 0;
        while (i < s.length()) {
            if (!isHostNameChar(s.charAt(i))) {
                return false;
            }
            ++i;
        }
        return !s.startsWith("-") && !s.endsWith("-");
    }
    
    private static boolean isHostNameChar(char c) {
        return c == '-' || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9');
    }
    
    private static boolean isDigit(char c) {
        return c >= '0' && c <= '9';
    }
    
    private static void escape(StringBuilder sb, String s) {
        for (int i = 0; i < s.length(); ++i) {
            int char1 = s.charAt(i);
            if (char1 == '.' || char1 == '\\') {
                sb.append('\\');
            }
            sb.append((char)char1);
        }
    }
    
    private static int compareLabels(String s, String s2) {
        int min = Math.min(s.length(), s2.length());
        int i = 0;
        while (i < min) {
            int char1 = s.charAt(i);
            int char2 = s2.charAt(i);
            if (char1 >= 'A' && char1 <= 'Z') {
                char1 = (char)(char1 + 32);
            }
            if (char2 >= 'A' && char2 <= 'Z') {
                char2 = (char)(char2 + 32);
            }
            if (char1 != char2) {
                return char1 - char2;
            }
            ++i;
        }
        return s.length() - s2.length();
    }
    
    private static String keyForLabel(String s) {
        StringBuffer sb = new StringBuffer(s.length());
        for (int i = 0; i < s.length(); ++i) {
            int char1 = s.charAt(i);
            if (char1 >= 'A' && char1 <= 'Z') {
                char1 = (char)(char1 + 32);
            }
            sb.append((char)char1);
        }
        return sb.toString();
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.writeObject(this.toString());
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        try {
            this.parse((String)objectInputStream.readObject());
        }
        catch (InvalidNameException ex) {
            throw new StreamCorruptedException("Invalid name: " + this.domain);
        }
    }
}
